import torch
import torch.nn as nn

class BaseNN(nn.Module):
    def __init__(self, **kwargs):
        super().__init__()
        assert type(self) != BaseNN, "BaseNN cannot be run directly"
        self.kwargs_checkpoint = kwargs.copy()
        self._initialize(**kwargs)
    
    def _initialize(self, **kwargs):
        """ initialize the model """
        pass

    def save(self, path):
        """ save model and kwargs to the same file """
        assert type(path) == str, "path must be a string"
        assert path[-4:] == ".pth" or path[-3:] == ".pt", "path must end with .pth or .pt"
        torch.save({
            "model_state_dict": self.state_dict(),
            "kwargs": self.kwargs_checkpoint
        }, path)

    def load(self, path, map_location=None):
        """ load model and kwargs from the same file """
        assert type(path) == str, "path must be a string"
        assert path[-4:] == ".pth" or path[-3:] == ".pt", "path must end with .pth or .pt"
        if map_location is None:
            checkpoint = torch.load(path)
        else:
            checkpoint = torch.load(path, map_location=map_location)
        self.kwargs_checkpoint = checkpoint["kwargs"]
        # add suppress_warnings to kwargs
        self.kwargs_checkpoint["suppress_warnings"] = True
        self._initialize(**self.kwargs_checkpoint)
        self.load_state_dict(checkpoint["model_state_dict"])

    def print_layers(self):
        """ print layer information """
        pass



import torch
import torch.nn as nn

from nn4n.model import BaseNN
from nn4n.layer import LinearLayer
from nn4n.utils import get_activation


class MLP(BaseNN):
    """
    Multi layer perceptron

    Keyword Arguments:
        - dims: dimensions of the network, default: [1, 100, 1]
        - preact_noise: noise added to pre-activation, default: 0
        - postact_noise: noise added to post-activation, default: 0
        - activation: activation function, default: "relu"
        - biases: use bias or not for each layer, a list of 3 values or a single value
            if a single value is passed, it will be broadcasted to a list of 3 values, it can be:
            - None: no bias
            - 'zero' or 0: bias initialized to 0
            - 'normal': bias initialized from a normal distribution
            - 'uniform': bias initialized from a uniform distribution
            if a list of 3 values is passed, each value can be either the same as above or
            a numpy array/torch tensor that directly specifies the bias
        - weights: distribution of weights for each layer, a list of 3 strings or
            a single string, if a single string is passed, it will be broadcasted to a list of 3 strings
            it can be:
            - 'normal': weights initialized from a normal distribution
            - 'uniform': weights initialized from a uniform distribution
            if a list of 3 values is passed, each string can be either the same as above or
            a numpy array/torch tensor that directly specifies the weights
        - sparsity_masks: use sparsity_masks or not, a list of 3 values or a single None
            if a single None is passed, it will be broadcasted to a list of 3 None
            if a list of 3 values is passed, each value can be either None or a numpy array/torch tensor
            that directly specifies the sparsity_masks
        - ei_masks: use ei_masks or not, a list of 3 values or a single None
            if a single None is passed, it will be broadcasted to a list of 3 None
            if a list of 3 values is passed, each value can be either None or a numpy array/torch tensor
            that directly specifies the ei_masks
        - plasticity_masks: use plasticity_masks or not, a list of 3 values or a single None
            if a single None is passed, it will be broadcasted to a list of 3 None
            if a list of 3 values is passed, each value can be either None or a numpy array/torch tensor
            that directly specifies the plasticity_masks
    """
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def _initialize(self, **kwargs):
        """ Initialize/Reinitialize the network """
        super()._initialize(**kwargs)
        # parameters that used in all layers
        # base parameters
        self.dims = kwargs.pop("dims", [10, 10])
        self.biases = kwargs.pop("biases", None)
        self.weights = kwargs.pop("weights", 'uniform')
        self.activation = get_activation(kwargs.pop("activation", "relu"))

        # network dynamics parameters
        self.sparsity_masks = kwargs.pop("sparsity_masks", None)
        self.ei_masks = kwargs.pop("ei_masks", None)
        self.plasticity_masks = kwargs.pop("plasticity_masks", None)

        # noise parameters
        self.preact_noise = kwargs.pop("preact_noise", 0)
        self.postact_noise = kwargs.pop("postact_noise", 0)

        # suppress warnings
        self.suppress_warnings = kwargs.pop("suppress_warnings", False)

        # check if all parameters meet the requirements
        if not self.suppress_warnings:
            self._handle_warnings(kwargs)
        self._check_parameters()

        self._init_layers()

    def _init_layers(self):
        self.layers = nn.ModuleList()
        for i in range(len(self.dims)-1):
            layer_struct = {
                "input_dim": self.dims[i],
                "output_dim": self.dims[i+1],
                "weights": self.weights[i],
                "biases": self.biases[i],
                "sparsity_mask": self.sparsity_masks[i],
                "ei_mask": self.ei_masks[i],
                "plasticity_mask": self.plasticity_masks[i],
            }
            l = LinearLayer(layer_struct=layer_struct)
            self.layers.append(l)

    def _handle_warnings(self, kwargs):
        """ Handle deprecated parameters """
        pass

    def _check_parameters(self):
        """ Check if the parameters meet the requirements """
        # check if dims is a list
        if not isinstance(self.dims, list):
            raise ValueError("dims must be a list of integers")
        n_weights = len(self.dims)

        self.biases = self._broadcast_values(self.biases, n_weights)
        self.weights = self._broadcast_values(self.weights, n_weights)
        self.sparsity_masks = self._broadcast_values(self.sparsity_masks, n_weights, is_mask=True)
        self.ei_masks = self._broadcast_values(self.ei_masks, n_weights, is_mask=True)
        self.plasticity_masks = self._broadcast_values(self.plasticity_masks, n_weights, is_mask=True)
        self.plasticity_masks = self._regularize_plas_masks(self.plasticity_masks, self.dims)
        self.preact_noise = self._broadcast_values(self.preact_noise, n_weights-1)  # no noise for the readout & input layer
        self.postact_noise = self._broadcast_values(self.postact_noise, n_weights-1)  # no noise for the readout & input layer

    def _regularize_plas_masks(self, masks, target_dim):
        if any(mask is not None for mask in masks):
            min_plas, max_plas = [], []
            for mask in masks:
                if mask is not None:
                    min_plas.append(mask.min())
                    max_plas.append(mask.max())
            min_plas, max_plas = min(min_plas), max(max_plas)
            if min_plas != max_plas:
                params = []
                for i in range(3):
                    if masks[i] is None: params.append(torch.ones(target_dim[i]))
                    else:
                        _temp_mask = (masks[i] - min_plas) / (max_plas - min_plas)
                        params.append(self._check_array(_temp_mask, "plasticity_masks", target_dim[i], i))
                # check the total number of unique plasticity values
                plasticity_scales = torch.unique(torch.cat([param.flatten() for param in params]))
                if len(plasticity_scales) > 5:
                    raise ValueError("The number of unique plasticity values cannot be larger than 5")
                return params
        return [torch.ones(target_dim[i]) for i in range(3)]

    def _standardize_masks(self, masks):
        assert masks is not None, "Masks cannot be None"
        if len(masks) != len(self.dims):
            raise ValueError("The length of the mask must be the same as the length of dims")

        for i in range(len(masks)):
            if masks[i] is None:
                continue
            if isinstance(masks[i], np.ndarray):
                masks[i] = torch.tensor(masks[i], dtype=torch.float32)
            if isinstance(masks[i], torch.Tensor):
                masks[i] = masks[i].float()
                # check shape is the same as dims
                if masks[i].shape != (self.dims[i], self.dims[i+1]):
                    raise ValueError(f"Mask shape mismatch, expected: {(self.dims[i], self.dims[i+1])}, got: {masks[i].shape}")
    
    def apply_plasticity(self):
        pass

    @staticmethod
    def _broadcast_values(value, length, is_mask=False):
        """ Broadcast a single value to a list if it's not already a list """
        if not isinstance(value, list):
            if is_mask and value != None:
                raise ValueError("Mask value cannot be auto-broadcasted")
            return [value] * length
        else:
            # check if the length of the list is correct
            if len(value) != length:
                raise ValueError(f"Expected a list of length {length}, got a list of length {len(value)}")
            return value

    def forward(self, x):
        """
        Inputs:
            - x: size=(batch_size, input_dim)
        """
        hidden_states = []
        for i, layer in enumerate(self.layers):
            if i == len(self.layers)-1:
                x = layer(x)
                break
            x = layer(x)
            x += torch.randn_like(x) * self.preact_noise[i]
            x = self.activation(x) + torch.randn_like(x) * self.postact_noise[i]
            hidden_states.append(x)
        
        return x, hidden_states

    def print_layers(self):
        """
        Print the layer information
        """
        for i, layer in enumerate(self.layers):
            layer.print_layers()

    def plot_layers(self):
        """
        Plot the layer information
        """
        for i, layer in enumerate(self.layers):
            layer.plot_layers()


import torch
import numpy as np
from scipy.stats import shapiro, chisquare


def to_numpy(w):
    if isinstance(w, np.ndarray):
        return w
    elif isinstance(w, torch.Tensor):
        return w.detach().cpu().numpy()
    else:
        raise ValueError("w must be either a torch.Tensor or a numpy.ndarray")


def self_connections(w):
    """ Check if w has self connections """
    return np.any(np.diag(w) != 0)


def determine_distribution(w):
    """
    Determine the distribution of w
    NOTE: assumed dist cannot be both uniform and normal
    """
    # randomly sample 5000 weights if w is too large
    w = w.flatten()
    if len(w) > 5000:
        w = np.random.choice(w, 5000)

    _, n = shapiro(w)
    counts, _ = np.histogram(w, bins=50)
    _, u = chisquare(counts)

    if n > u:
        return "normal"
    else:
        return "uniform"


def check_activation(act, act_func):
    print(act)
    pos = torch.tensor([1, 2, 3, 4, 5])
    neg = torch.tensor([-1, -2, -3, -4, -5])
    if act == 'relu':
        zero = torch.tensor([0, 0, 0, 0, 0])
        return torch.all(act_func(pos) == pos) and torch.all(act_func(neg) == zero)
    elif act == 'tanh':
        pos_ans = torch.tensor([0.7616, 0.9640, 0.9951, 0.9993, 0.9999])
        neg_ans = torch.tensor([-0.7616, -0.9640, -0.9951, -0.9993, -0.9999])
        return torch.allclose(act_func(pos), pos_ans, atol=1e-4) and torch.allclose(act_func(neg), neg_ans, atol=1e-4)
    elif act == 'sigmoid':
        pos_ans = torch.tensor([0.7311, 0.8808, 0.9526, 0.9820, 0.9933])
        neg_ans = torch.tensor([0.2689, 0.1192, 0.0474, 0.0180, 0.0067])
        return torch.allclose(act_func(pos), pos_ans, atol=1e-4) and torch.allclose(act_func(neg), neg_ans, atol=1e-4)
    else:
        return False


def get_weight_info(w):
    """ Check the distribution of w and get some statistics """
    w = to_numpy(w)
    specs = {
        "mean": np.mean(w),
        "std": np.std(w),
        "min": np.min(w),
        "max": np.max(w),
        "dist": determine_distribution(w),
        "self_connections": self_connections(w),
    }
    if np.any(np.isnan(w)):
        specs["contains_nan"] = "nan"

    if np.any(np.isinf(w)):
        specs["contains_inf"] = "inf"

    return specs



import numpy as np
import nn4n.utils as utils

class BaseMask():
    """ Base class for all masks """
    def __init__(self, **kwargs):
        """
        @kwarg hidden_size: number of hidden neurons in total, must be defined
        @kwarg input_dim: input dimension, default: 1
        @kwarg output_dim: output dimension, default: 1
        """
        self.dims = kwargs.get("dims", [1, 100, 1])
        self.hidden_size = self.dims[1]
        self.input_dim = self.dims[0]
        self.output_dim = self.dims[2]

        # cannot be run as a child class
        assert self.__class__.__name__ != "BaseMask", "BaseMask cannot be run as a child class"

    def _check_parameters(self):
        """
        Check if parameters are valid
        """
        # check hidden_size
        assert self.hidden_size is not None, "hidden_size must be defined"
        assert isinstance(self.hidden_size, int), "hidden_size must be int"
        assert self.hidden_size > 0, "hidden_size must be positive"

        # check input_dim
        assert self.input_dim is not None, "input_dim must be defined"
        assert isinstance(self.input_dim, int), "input_dim must be int"
        assert self.input_dim > 0, "input_dim must be positive"

        # check output_dim
        assert self.output_dim is not None, "output_dim must be defined"
        assert isinstance(self.output_dim, int), "output_dim must be int"
        assert self.output_dim > 0, "output_dim must be positive"

    def _generate_mask(self):
        """
        Generate the mask for the multi-area network
        """
        self._generate_hidden_mask()
        self._generate_input_mask()
        self._generate_readout_mask()

    def _generate_hidden_mask(self):
        """
        Generate the mask for the hidden layer
        """
        raise NotImplementedError

    def _generate_input_mask(self):
        """
        Generate the mask for the input layer
        """
        raise NotImplementedError

    def _generate_readout_mask(self):
        """
        Generate the mask for the readout layer
        """
        raise NotImplementedError

    def get_input_idx(self):
        """
        Return the indices of neurons that receive input
        """
        raise NotImplementedError

    def get_non_input_idx(self):
        """
        Return the indices of neurons that do not receive input
        """
        raise NotImplementedError

    def get_readout_idx(self):
        """ Return the indices of neurons that send output """
        raise NotImplementedError

    def get_areas(self):
        """ Return the number of areas """
        raise NotImplementedError

    def get_area_idx(self, area):
        """ Return the indices of neurons in area """
        raise NotImplementedError

    def _generate_sparse_matrix(self, n, m, p):
        """
        Generate a sparse matrix with size n x m and density p. 1 if connection exists, 0 otherwise
        """
        assert 0 <= p <= 1, "p must be between 0 and 1"
        mask = np.random.rand(n, m) < p
        return mask.astype(int)

    def plot_masks(self):
        if self.input_mask is not None:
            input_mask_ = self.input_mask if self.input_mask.shape[1] > self.input_mask.shape[0] else self.input_mask.T
            utils.plot_connectivity_matrix(input_mask_, "Input Layer Mask", False)

        if self.hidden_mask is not None:
            utils.plot_connectivity_matrix(self.hidden_mask, "Hidden Layer Mask", False)

        if self.readout_mask is not None:
            readout_mask_ = self.readout_mask if self.readout_mask.shape[1] > self.readout_mask.shape[0] else self.readout_mask.T
            utils.plot_connectivity_matrix(readout_mask_, "Readout Layer Mask", False)

    def get_masks(self):
        """ Return the masks """
        assert self.input_mask is not None, "input_mask is not generated"
        assert self.hidden_mask is not None, "hidden_mask is not generated"
        assert self.readout_mask is not None, "readout_mask is not generated"
        return [self.input_mask.T, self.hidden_mask.T, self.readout_mask.T]



def weight_check(W):
    # must be torch tensor
    assert isinstance(W, torch.Tensor), "W must be a torch tensor"

def frobenius_norm(W, avg=True):
    weight_check(W)
    return T.sqrt(T.sum(W**2))/W.numel() if avg else T.sqrt(T.sum(W**2))

def l1_norm(W, avg=True):
    weight_check(W)
    return T.sum(T.abs(W))/W.numel() if avg else T.sum(T.abs(W))



import torch
import torch.nn as nn

import numpy as np
import nn4n.utils as utils


class HiddenLayer(nn.Module):
    """
    Hidden layer of the RNN. The layer is initialized by passing specs in layer_struct.
    
    Required keywords in layer_struct:
        - input_dim: input dimension
        - output_dim: output dimension
        - weights: weight matrix init method/init weight matrix
        - biases: bias vector init method/init bias vector
        - sparsity_mask: mask for sparse connectivity
        - ei_mask: mask for Dale's law
        - plasticity_mask: mask for plasticity
    """
    def __init__(self, layer_struct, **kwargs):
        super().__init__()
        # some params are for verbose printing
        self.input_dim = layer_struct['input_dim']
        self.output_dim = layer_struct['output_dim']
        self.ei_mask = layer_struct['ei_mask'].T if layer_struct['ei_mask'] is not None else None
        self.sparsity_mask = layer_struct['sparsity_mask'].T if layer_struct['sparsity_mask'] is not None else None
        self.plasticity_mask = layer_struct['plasticity_mask'].T if layer_struct['plasticity_mask'] is not None else None
        self.plasticity_scales = torch.unique(self.plasticity_mask) # all unique plasticity values in the plasticity mask
        
        # generate weights and bias
        self.weight = self._generate_weight(layer_struct['weights'])
        self.bias = self._generate_bias(layer_struct['biases'])
        
        # initialize constraints
        self._init_constraints()

        # parameterize the weights and bias
        self.weight = nn.Parameter(self.weight, requires_grad=layer_struct['weights'] is not None)
        self.bias = nn.Parameter(self.bias, requires_grad=layer_struct['biases'] is not None)

    # INITIALIZATION
    # ======================================================================================
    def _generate_weight(self, weight_init):
        """ Generate random weight """
        if weight_init == 'uniform':
            # if uniform, let w be uniform in [-sqrt(k), sqrt(k)]
            sqrt_k = torch.sqrt(torch.tensor(1/self.input_dim))
            w = torch.rand(self.output_dim, self.input_dim) * sqrt_k
            w = w * 2 - sqrt_k
        elif weight_init == 'normal':
            w = torch.randn(self.output_dim, self.input_dim) / torch.sqrt(torch.tensor(self.input_dim))
        elif weight_init == 'zero':
            w = torch.zeros((self.output_dim, self.input_dim))
        elif type(weight_init) == np.ndarray:
            w = torch.from_numpy(weight_init)
        else:
            raise NotImplementedError
        return w.float()

    def _generate_bias(self, bias_init):
        """ Generate random bias """
        if bias_init == 'uniform':
            # if uniform, let b be uniform in [-sqrt(k), sqrt(k)]
            sqrt_k = torch.sqrt(torch.tensor(1/self.input_dim))
            b = torch.rand(self.output_dim) * sqrt_k
            b = b * 2 - sqrt_k
        elif bias_init == 'normal':
            b = torch.randn(self.output_dim) / torch.sqrt(torch.tensor(self.input_dim))
        elif bias_init == 'zero' or bias_init == None:
            b = torch.zeros(self.output_dim)
        elif type(bias_init) == np.ndarray:
            b = torch.from_numpy(bias_init)
        else:
            raise NotImplementedError
        return b.float()

    def _init_constraints(self):
        """
        Initialize constraints
        It will also balance excitatory and inhibitory neurons
        """
        if self.sparsity_mask is not None:
            self.weight *= self.sparsity_mask
        if self.ei_mask is not None:
            self.weight[self.ei_mask == 1] = torch.clamp(self.weight[self.ei_mask == 1], min=0)
            self.weight[self.ei_mask == -1] = torch.clamp(self.weight[self.ei_mask == -1], max=0)
            self._balance_excitatory_inhibitory()

    def _balance_excitatory_inhibitory(self):
        """ Balance excitatory and inhibitory weights """
        scale_mat = torch.ones_like(self.weight)
        ext_sum = self.weight[self.sparsity_mask == 1].sum()
        inh_sum = self.weight[self.sparsity_mask == -1].sum()
        if ext_sum == 0 or inh_sum == 0:
            # automatically stop balancing if one of the sums is 0
            # devide by 10 to avoid recurrent explosion/decay
            self.weight /= 10
        else:
            if ext_sum > abs(inh_sum):
                _scale = abs(inh_sum).item() / ext_sum.item()
                scale_mat[self.sparsity_mask == 1] = _scale
            elif ext_sum < abs(inh_sum):
                _scale = ext_sum.item() / abs(inh_sum).item()
                scale_mat[self.sparsity_mask == -1] = _scale
            # apply scaling
            self.weight *= scale_mat
    # ======================================================================================

    # FORWARD
    # ======================================================================================
    def to(self, device):
        """ Move the network to the device (cpu/gpu) """
        super().to(device)
        if self.sparsity_mask is not None:
            self.sparsity_mask = self.sparsity_mask.to(device)
        if self.ei_mask is not None:
            self.ei_mask = self.ei_mask.to(device)
        if self.bias.requires_grad:
            self.bias = self.bias.to(device)

    def forward(self, x):
        """ 
        Forwardly update network

        Inputs:
            - x: input, shape: (batch_size, input_dim)

        Returns:
            - state: shape: (batch_size, hidden_size)
        """
        return x.float() @ self.weight.T + self.bias

    def apply_plasticity(self):
        """
        Apply plasticity mask to the weight gradient
        """
        with torch.no_grad():
            # assume the plasticity mask are all valid and being checked in ctrnn class
            for scale in self.plasticity_scales:
                if self.weight.grad is not None:
                    self.weight.grad[self.plasticity_mask == scale] *= scale
                else:
                    raise RuntimeError("Weight gradient is None, possibly because the forward loop is non-differentiable")

    def _enforce_spec_rad(self):
        """ Enforce spectral radius """
        print('WARNING: spectral radius not applied, the feature is deprecated, use scaling instead')

    def enforce_constraints(self):
        """ 
        Enforce sparsity and excitatory/inhibitory constraints if applicable.
        This is by default automatically called after each forward pass,
        but can be called manually if needed
        """
        if self.sparsity_mask is not None:
            self._enforce_sparsity()
        if self.ei_mask is not None:
            self._enforce_ei()

    def _enforce_sparsity(self):
        """ Enforce sparsity """
        w = self.weight.detach().clone() * self.sparsity_mask
        self.weight.data.copy_(torch.nn.Parameter(w))

    def _enforce_ei(self):
        """ Enforce Dale's law """
        w = self.weight.detach().clone()
        w[self.ei_mask == 1] = torch.clamp(w[self.ei_mask == 1], min=0)
        w[self.ei_mask == -1] = torch.clamp(w[self.ei_mask == -1], max=0)
        self.weight.data.copy_(torch.nn.Parameter(w))
    # ======================================================================================

    # HELPER FUNCTIONS
    # ======================================================================================
    def get_weight(self):
        """ Get the value of weight """ 
        

    def set_weight(self, weight):
        """ Set the value of weight """
        assert weight.shape == self.weight.shape, f"Weight shape mismatch, expected {self.weight.shape}, got {weight.shape}"
        with torch.no_grad():
            self.weight.copy_(weight)

    def plot_layers(self):
        """ Plot the weights matrix and distribution of each layer """
        weight = self.weight.cpu() if self.weight.device != torch.device('cpu') else self.weight
        utils.plot_connectivity_matrix_dist(weight.detach().numpy(), "Hidden Layer", False, self.sparsity_mask is not None)
    
    def print_layers(self):
        """ Print the specs of each layer """
        param_dict = {
            "input_dim": self.input_dim,
            "output_dim": self.output_dim,
            "weight_learnable": self.weight.requires_grad,
            "weight_min": self.weight.min().item(),
            "weight_max": self.weight.max().item(),
            "bias_learnable": self.bias.requires_grad,
            "bias_min": self.bias.min().item(),
            "bias_max": self.bias.max().item(),
            "sparsity": self.sparsity_mask.sum() / self.sparsity_mask.numel() if self.sparsity_mask is not None else 1,
            # "spectral_radius": torch.abs(torch.linalg.eig(self.weight)[0]).max().item(),
        }
        utils.print_dict("Hidden Layer", param_dict)
        # utils.plot_eigenvalues(weight.detach().numpy(), "Hidden Layer")
    # ======================================================================================



import torch
import numpy as np
import matplotlib.pyplot as plt


def print_dict(title, params):
    """
    Print a dictionary in key-value pairs with a nice format

    Inputs:
        - title: title of the dictionary
        - params: dictionary to be printed
    """
    print(f"{title}: ")
    maxlen = max([len(s) for s in params.keys()])
    for k in params.keys():
        print(3*' ' + '| {}:{}{}'.format(k, (maxlen - len(k) + 1)*' ', params[k]))
    print()


def get_activation(act):
    """
    Return the activation function given the name

    Inputs:
        - act: name of the activation function (supported values: "relu", "tanh", "sigmoid", "retanh")
    Returns:
        - activation function
    """
    if act == "relu":
        return torch.relu
    elif act == "tanh":
        return torch.tanh
    elif act == "sigmoid":
        return torch.sigmoid
    elif act == "retanh":
        return lambda x: torch.maximum(torch.tanh(x), torch.tensor(0))
    else:
        raise NotImplementedError


def plot_connectivity_matrix_dist(w, title, colorbar=True, ignore_zeros=False):
    """
    Plot the distribution of a connectivity matrix

    Inputs:
        - w: connectivity matrix, must be a numpy array or a torch tensor
        - title: title of the plot
        - colorbar: whether to show the colorbar (default: True)
        - ignore_zeros: whether to ignore zeros in the distribution (needed for sparse matrices) (default: False)
    """
    if type(w) == torch.Tensor:
        w = w.detach().numpy()
    
    r = np.max(np.abs(w))

    img_width, hist_height = 6, 2
    hw_ratio = w.shape[0] / w.shape[1]
    plt_height = img_width * hw_ratio + hist_height

    fig, ax = plt.subplots(figsize=(img_width, plt_height))
    ax.imshow(-w, cmap='bwr', vmin=-r, vmax=r)
    ax.set_title(f'{title}' if not ignore_zeros else f'{title} (nonzero)')
    if colorbar:
        fig.colorbar(ax.imshow(-w, cmap='bwr', vmin=-r, vmax=r), ax=ax)
    if w.shape[1] < 5:
        ax.set_xticks([])
    if w.shape[0] < 5:
        ax.set_yticks([])
    plt.tight_layout()
    plt.show()

    fig, ax = plt.subplots(figsize=(img_width, hist_height))
    ax.set_title(f'{title} distribution')
    if ignore_zeros:
        mean_nonzero = np.mean(np.abs(w)[np.abs(w) != 0])
        ax.hist(w[np.abs(w) > mean_nonzero*0.001].flatten(), bins=100)
    else:
        ax.hist(w.flatten(), bins=100)
    plt.tight_layout()
    plt.show()


def plot_connectivity_matrix(w, title, colorbar=True):
    """
    Plot a connectivity matrix

    Inputs:
        - w: connectivity matrix, must be a numpy array or a torch tensor
        - title: title of the plot
        - colorbar: whether to show the colorbar (default: True)
    """
    if type(w) == torch.Tensor:
        w = w.detach().numpy()

    r = np.max(np.abs(w))

    fig, ax = plt.subplots(figsize=(6, 6*w.shape[0]/w.shape[1]))

    ax.imshow(-w, cmap='bwr', vmin=-r, vmax=r)
    ax.set_title(title)
    if colorbar:
        fig.colorbar(ax.imshow(w, cmap='bwr', vmin=-r, vmax=r), ax=ax)
    # plt.tight_layout()
    plt.show()


def plot_eigenvalues(w, title):
    """
    Plot the eigenvalues of a connectivity matrix in the complex plane

    Inputs:
        - w: connectivity matrix, must be a numpy array or a torch tensor
        - title: title of the plot
    """
    if type(w) == torch.Tensor:
        w = w.detach().numpy()

    eigvals = np.linalg.eigvals(w)
    plt.figure(figsize=(4, 4))
    plt.plot(eigvals.real, eigvals.imag, 'o')
    max_eig = np.round(np.max(np.abs(eigvals)), 1)
    plt.xticks(np.arange(-max_eig, max_eig, 0.5), rotation=45)
    plt.yticks(np.arange(-max_eig, max_eig, 0.5))
    plt.xlabel(r'$\lambda_{real}$')
    plt.ylabel(r'$\lambda_{imag}$')
    plt.title(title)
    plt.show()



import numpy as np
from nn4n.mask.base_mask import BaseMask

class RandomInput(BaseMask):
    def __init__(self, **kwargs):
        """
        Generate a network with randomly injected input and readout neurons
        @kwarg input_spar: sparsity of the input layer, default: 0.1
        @kwarg readout_spar: sparsity of the readout layer, default: 0.1
        @kwarg overlap: overlap between input and readout neurons, default: True
        """
        super().__init__(**kwargs)
        self.input_spar = kwargs.get("input_spar", 1)
        self.readout_spar = kwargs.get("readout_spar", 1)
        self.hidden_spar = kwargs.get("hidden_spar", 1)
        self.overlap = kwargs.get("overlap", True)

        # run if it is not a child class
        if self.__class__.__name__ == "RandomInput":
            self._check_parameters()
            self._generate_mask()

    def _check_parameters(self):
        """
        Check if parameters are valid
        """
        super()._check_parameters()

        # check input_spar
        assert isinstance(self.input_spar, float), "input_spar must be float"
        assert 0 <= self.input_spar <= 1, "input_spar must be between 0 and 1"

        # check readout_spar
        assert isinstance(self.readout_spar, float), "readout_spar must be float"
        assert 0 <= self.readout_spar <= 1, "readout_spar must be between 0 and 1"

        # if not overlap, check if input_spar + readout_spar <= 1
        if not self.overlap:
            assert self.input_spar + self.readout_spar <= 1, "input_spar + readout_spar must be less than 1 if overlap is True"

    def _generate_hidden_mask(self):
        """
        Generate the mask for the hidden layer
        """
        self.hidden_mask = np.random.uniform(0, 1, (self.hidden_size, self.hidden_size))
        self.hidden_mask[self.hidden_mask > self.hidden_spar] = 0
        self.hidden_mask[self.hidden_mask > 0] = 1

    def _generate_input_mask(self):
        """
        Generate the mask for the input layer
        """
        self.input_idx = np.random.choice(self.hidden_size, int(self.hidden_size * self.input_spar), replace=False)
        self.non_input_idx = np.setdiff1d(np.arange(self.hidden_size), self.input_idx)
        self.input_mask = np.zeros((self.hidden_size, self.input_dim))
        self.input_mask[np.ix_(self.input_idx, np.arange(self.input_dim))] = self._generate_sparse_matrix(len(self.input_idx), self.input_dim, 1)

    def _generate_readout_mask(self):
        """
        Generate the mask for the readout layer
        """
        if self.overlap:
            self.readout_idx = np.random.choice(self.hidden_size, int(self.hidden_size * self.readout_spar), replace=False)
        else:
            self.readout_idx = np.random.choice(self.non_input_idx, int(self.hidden_size * self.readout_spar), replace=False)

        self.readout_mask = np.zeros((self.output_dim, self.hidden_size))
        self.readout_mask[np.ix_(np.arange(self.output_dim), self.readout_idx)] = \
            self._generate_sparse_matrix(self.output_dim, len(self.readout_idx), 1)

    def get_input_idx(self):
        """
        Return the indices of neurons that receive input
        """
        return self.input_idx

    def get_non_input_idx(self):
        """
        Return the indices of neurons that do not receive input
        """
        return self.non_input_idx

    def get_readout_idx(self):
        """
        Return the indices of neurons that send output
        """
        return self.readout_idx

    def get_input_mask(self):
        """
        Return the mask for the input layer
        """
        return self.input_mask

    def get_areas(self):
        """
        Return the area of each layer
        """
        return ['area_input', 'area_readout']

    def get_area_idx(self, area):
        """
        Return the indices of neurons in each area
        """
        indicies = [self.input_idx, self.readout_idx]
        if isinstance(area, str):
            area = self.get_areas().index(area)
        return indicies[area]



import numpy as np
from nn4n.mask.base_mask import BaseMask

class MultiArea(BaseMask):
    def __init__(self, **kwargs):
        """
        Generate the multi-area network mask
        @kwarg n_areas: number of areas, list or int, default: 2
        @kwarg area_connectivities: connectivity between areas, list or np.ndarray, default: [0.1, 0.1]
        """
        super().__init__(**kwargs)
        self.n_areas = kwargs.get("n_areas", 2)
        self.area_connectivities = kwargs.get("area_connectivities", [0.1, 0.1])
        self.input_areas = kwargs.get("input_areas", None)
        self.readout_areas = kwargs.get("readout_areas", None)

        # run if it is not a child class
        if self.__class__.__name__ == "MultiArea":
            self._check_parameters()
            self._generate_mask()

    def _check_parameters(self):
        """ Check if parameters are valid """
        super()._check_parameters()

        # check n_areas
        if isinstance(self.n_areas, int):
            assert self.hidden_size % self.n_areas == 0, "hidden_size must be devideable by n_areas"
            # create a node assignment list
            node_assigment = np.zeros(self.n_areas, dtype=np.ndarray)
            for i in range(self.n_areas):
                node_assigment[i] = np.arange(self.hidden_size // self.n_areas) + i * (self.hidden_size // self.n_areas)
            self.node_assigment = node_assigment
        elif isinstance(self.n_areas, list):
            assert sum(self.n_areas) == self.hidden_size, "sum of n_areas must be equal to hidden_size"
            assert np.all(np.array(self.n_areas) > 0), "elements of n_areas must be larger than 0"
            # create a node assignment list
            node_assigment = np.zeros(len(self.n_areas), dtype=np.ndarray)
            for i in range(len(self.n_areas)):
                node_assigment[i] = np.arange(self.n_areas[i]) + sum(self.n_areas[:i])
            self.node_assigment = node_assigment
            self.n_areas = len(self.n_areas)
        else:
            assert False, "n_areas must be int or list"

        if self.n_areas == 1:
            self.input_areas = np.array([0])
            self.readout_areas = np.array([0])

        # check area_connectivities
        assert np.all(0 <= np.array(self.area_connectivities)) and np.all(np.array(self.area_connectivities) <= 1), \
            "area_connectivities must be between 0 and 1"
        if isinstance(self.area_connectivities, list):
            assert len(self.area_connectivities) in [2, 3], "length of area_connectivities must be 2 or 3"
            # transform list to np.ndarray
            area_connectivities = np.zeros((self.n_areas, self.n_areas))
            area_connectivities[np.tril_indices(self.n_areas, -1)] = self.area_connectivities[0]
            area_connectivities[np.triu_indices(self.n_areas, 1)] = self.area_connectivities[1]
            area_connectivities[np.diag_indices(self.n_areas)] = self.area_connectivities[2] if len(self.area_connectivities) == 3 else 1
            self.area_connectivities = area_connectivities
        elif isinstance(self.area_connectivities, np.ndarray):
            assert self.area_connectivities.shape == (self.n_areas, self.n_areas), "shape of area_connectivities must be (n_areas, n_areas)"
        else:
            assert False, "area_connectivities must be list or np.ndarray"

        # check input_areas and readout_areas
        if self.input_areas is None:
            self.input_areas = np.arange(self.n_areas)
        else:
            self.input_areas = np.array(self.input_areas)
            assert np.all(0 <= self.input_areas) and np.all(self.input_areas < self.n_areas), "input_areas must be between 0 and n_areas"
        if self.readout_areas is None:
            self.readout_areas = np.arange(self.n_areas)
        else:
            self.readout_areas = np.array(self.readout_areas)
            assert np.all(0 <= self.readout_areas) and np.all(self.readout_areas < self.n_areas), "readout_areas must be between 0 and n_areas"

    def _generate_hidden_mask(self):
        """ Generate the mask for the hidden layer """
        hidden_mask = np.zeros((self.hidden_size, self.hidden_size))
        for i in range(self.n_areas):
            for j in range(self.n_areas):
                if self.area_connectivities[i, j] > 0:
                    area_i_size = len(self.node_assigment[i])
                    area_j_size = len(self.node_assigment[j])
                    hidden_mask[np.ix_(self.node_assigment[i], self.node_assigment[j])] = \
                        self._generate_sparse_matrix(area_i_size, area_j_size, self.area_connectivities[i, j])
        self.hidden_mask = hidden_mask

    def _generate_input_mask(self):
        """ Generate the mask for the input layer """
        input_mask = np.zeros((self.hidden_size, self.input_dim))
        for i in self.input_areas:
            area_i_size = len(self.node_assigment[i])
            input_mask[np.ix_(self.node_assigment[i], np.arange(self.input_dim))] = self._generate_sparse_matrix(area_i_size, self.input_dim, 1)
        self.input_mask = input_mask

    def _generate_readout_mask(self):
        """ Generate the mask for the readout layer """
        readout_mask = np.zeros((self.output_dim, self.hidden_size))
        for i in self.readout_areas:
            area_i_size = len(self.node_assigment[i])
            readout_mask[np.ix_(np.arange(self.output_dim), self.node_assigment[i])] = self._generate_sparse_matrix(self.output_dim, area_i_size, 1)
        self.readout_mask = readout_mask

    def get_input_idx(self):
        """ Return the indices of neurons that receive input """
        return np.where(self.input_mask.sum(axis=1) != 0)[0]

    def get_non_input_idx(self):
        """ Return the indices of neurons that do not receive input """
        return np.where(self.input_mask.sum(axis=1) == 0)[0]

    def get_readout_idx(self):
        """ Return the indices of neurons that send output """
        return np.where(self.readout_mask.sum(axis=0) != 0)[0]

    def get_area_idx(self, area):
        """ Return the indices of neurons in area """
        if isinstance(area, str):
            area = self.get_areas().index(area)
        return self.node_assigment[area]

    def get_areas(self):
        """ Return the number of areas """
        return [f"area_{i+1}" for i in range(self.n_areas)]



import torch
import torch.nn as nn
from nn4n.utils import print_dict, get_activation

from nn4n.layer import HiddenLayer
from nn4n.layer import LinearLayer


class RecurrentLayer(nn.Module):
    def __init__(
            self,
            hidden_size,
            positivity_constraints,
            sparsity_constraints,
            layer_distributions,
            layer_biases,
            layer_masks,
            preact_noise,
            postact_noise,
            learnable=True,
            **kwargs
            ):
        """
        Hidden layer of the RNN
        Parameters:
            @param hidden_size: number of hidden neurons
            @param positivity_constraints: whether to enforce positivity constraint
            @param sparsity_constraints: use sparsity_constraints or not
            @param layer_distributions: distribution of weights for each layer, a list of 3 strings
            @param layer_biases: use bias or not for each layer, a list of 3 boolean values

        Keyword Arguments:
            @kwarg activation: activation function, default: "relu"
            @kwarg preact_noise: noise added to pre-activation, default: 0
            @kwarg postact_noise: noise added to post-activation, default: 0
            @kwarg dt: time step, default: 1
            @kwarg tau: time constant, default: 1

            @kwarg input_dim: input dimension, default: 1

            @kwarg hidden_dist: distribution of hidden layer weights, default: "normal"
            @kwarg self_connections: allow self connections or not, default: False
            @kwarg init_state: initial state of the network, 'zero', 'keep', or 'learn'
        """
        super().__init__()

        self.hidden_size = hidden_size
        self.preact_noise = preact_noise
        self.postact_noise = postact_noise
        self.alpha = kwargs.get("dt", 10) / kwargs.get("tau", 100)
        self.layer_distributions = layer_distributions
        self.layer_biases = layer_biases
        self.layer_masks = layer_masks
        self.hidden_state = torch.zeros(self.hidden_size)
        self.init_state = kwargs.get("init_state", 'zero')
        self.act = kwargs.get("activation", "relu")
        self.activation = get_activation(self.act)
        self._set_hidden_state()

        self.input_layer = LinearLayer(
            positivity_constraints=positivity_constraints[0],
            sparsity_constraints=sparsity_constraints[0],
            output_dim=self.hidden_size,
            input_dim=kwargs.pop("input_dim", 1),
            use_bias=self.layer_biases[0],
            dist=self.layer_distributions[0],
            mask=self.layer_masks[0],
            learnable=learnable[0],
        )
        self.hidden_layer = HiddenLayer(
            hidden_size=self.hidden_size,
            sparsity_constraints=sparsity_constraints[1],
            positivity_constraints=positivity_constraints[1],
            dist=self.layer_distributions[1],
            use_bias=self.layer_biases[1],
            scaling=kwargs.get("scaling", 1.0),
            mask=self.layer_masks[1],
            self_connections=kwargs.get("self_connections", False),
            learnable=learnable[1],
        )

    # INITIALIZATION
    # ==================================================================================================
    def _set_hidden_state(self):
        """
        Add the hidden layer to the parameter
        """
        if self.init_state == 'learn':
            self.hidden_state = torch.nn.Parameter(torch.zeros(self.hidden_size), requires_grad=True)
        else:
            self.hidden_state = torch.nn.Parameter(torch.zeros(self.hidden_size), requires_grad=False)
    # ==================================================================================================

    # FORWARD
    # ==================================================================================================
    def _reset_state(self):
        if self.init_state == 'learn' or self.init_state == 'keep':
            return self.hidden_state
        else:
            return torch.zeros(self.hidden_size)

    def enforce_constraints(self):
        self.input_layer.enforce_constraints()
        self.hidden_layer.enforce_constraints()

    def recurrence(self, fr_t, v_t, u_t):
        """ Recurrence function """
        # through input layer
        v_in_u_t = self.input_layer(u_t)  # u_t @ W_in
        # through hidden layer
        v_hid_fr_t = self.hidden_layer(fr_t)  # fr_t @ W_hid + b
        # update hidden state
        v_t = (1-self.alpha)*v_t + self.alpha*(v_hid_fr_t+v_in_u_t)

        # add pre-activation noise
        if self.preact_noise > 0:
            preact_epsilon = torch.randn((u_t.size(0), self.hidden_size), device=u_t.device) * self.preact_noise
            v_t = v_t + self.alpha*preact_epsilon
    
        # apply activation function
        fr_t = self.activation(v_t)
    
        # add post-activation noise
        if self.postact_noise > 0:
            postact_epsilon = torch.randn((u_t.size(0), self.hidden_size), device=u_t.device) * self.postact_noise
            fr_t = fr_t + postact_epsilon
        
        return fr_t, v_t

    def forward(self, input):
        """
        Propogate input through the network.
        @param input: shape=(seq_len, batch, input_dim), network input
        @return stacked_states: shape=(seq_len, batch, hidden_size), stack of hidden layer status
        """
        v_t = self._reset_state().to(input.device)
        fr_t = self.activation(v_t)
        # update hidden state and append to stacked_states
        stacked_states = []
        for i in range(input.size(0)):
            fr_t, v_t = self.recurrence(fr_t, v_t, input[i])
            # append to stacked_states
            stacked_states.append(fr_t)

        # if keeping the last state, save it to hidden_state
        if self.init_state == 'keep':
            self.hidden_state = fr_t.detach().clone()  # TODO: haven't tested this yet

        return torch.stack(stacked_states, dim=0)
    # ==================================================================================================

    # HELPER FUNCTIONS
    # ==================================================================================================
    def to(self, device):
        """
        Move the network to the device (cpu/gpu)
        """
        super().to(device)
        self.input_layer.to(device)
        self.hidden_layer.to(device)
        self.hidden_state = self.hidden_state.to(device)

    def print_layers(self):
        param_dict = {
            "hidden_min": self.hidden_state.min(),
            "hidden_max": self.hidden_state.max(),
            "hidden_mean": self.hidden_state.mean(),
            "preact_noise": self.preact_noise,
            "postact_noise": self.postact_noise,
            "activation": self.act,
            "alpha": self.alpha,
        }
        self.input_layer.print_layers()
        print_dict("Recurrence", param_dict)
        self.hidden_layer.print_layers()

    def plot_layers(self, **kwargs):
        self.input_layer.plot_layers()
        self.hidden_layer.plot_layers()
    # ==================================================================================================



import torch
import torch.nn as nn
import numpy as np
import warnings

from nn4n.model import BaseNN
from nn4n.layer import RecurrentLayer
from nn4n.layer import LinearLayer

param_list = ['dims', 'preact_noise', 'postact_noise', 
              'init_state', 'activation', 'tau', 'dt', 'weights',
              'biases', 'sparsity_masks', 'ei_masks', 'plasticity_masks']

dep_param_list = ['input_dim', 'output_dim', 'hidden_size', 'ei_balance', 'learnable',
              'allow_negative', 'use_dale', 'new_synapses', 'positivity_constraints', 
              'sparsity_constraints', 'layer_distributions', 'layer_biases', 
              'layer_masks', 'scaling', 'self_connections']

class CTRNN(BaseNN):
    """
    Recurrent network model

    Keyword Arguments:
        - dims: dimensions of the network, default: [1, 100, 1]
        - preact_noise: noise added to pre-activation, default: 0
        - postact_noise: noise added to post-activation, default: 0
        - init_state: initial state of the network. It defines the hidden state at t=0.
        - activation: activation function, default: "relu", can be "relu", "sigmoid", "tanh", "retanh"
        - dt: time step, default: 10
        - tau: time constant, default: 100
        - biases: use bias or not for each layer, a list of 3 values or a single value
            if a single value is passed, it will be broadcasted to a list of 3 values, it can be:
            - None: no bias
            - 'zero' or 0: bias initialized to 0
            - 'normal': bias initialized from a normal distribution
            - 'uniform': bias initialized from a uniform distribution
            if a list of 3 values is passed, each value can be either the same as above or
            a numpy array/torch tensor that directly specifies the bias
        - weights: distribution of weights for each layer, a list of 3 strings or
            a single string, if a single string is passed, it will be broadcasted to a list of 3 strings
            it can be:
            - 'normal': weights initialized from a normal distribution
            - 'uniform': weights initialized from a uniform distribution
            if a list of 3 values is passed, each string can be either the same as above or
            a numpy array/torch tensor that directly specifies the weights
        - sparsity_masks: use sparsity_masks or not, a list of 3 values or a single None
            if a single None is passed, it will be broadcasted to a list of 3 None
            if a list of 3 values is passed, each value can be either None or a numpy array/torch tensor
            that directly specifies the sparsity_masks
        - ei_masks: use ei_masks or not, a list of 3 values or a single None
            if a single None is passed, it will be broadcasted to a list of 3 None
            if a list of 3 values is passed, each value can be either None or a numpy array/torch tensor
            that directly specifies the ei_masks
        - plasticity_masks: use plasticity_masks or not, a list of 3 values or a single None
            if a single None is passed, it will be broadcasted to a list of 3 None
            if a list of 3 values is passed, each value can be either None or a numpy array/torch tensor
            that directly specifies the plasticity_masks
        - synapse_growth_masks: use synapse_growth_masks or not, a list of 3 values or a single None
            if a single None is passed, it will be broadcasted to a list of 3 None
            if a list of 3 values is passed, each value can be either None or a numpy array/torch tensor
            that directly specifies the probability of growing a synapse at the selected location if there is no synapse.
    """
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    # INITIALIZATION
    # ======================================================================================
    def _initialize(self, **kwargs):
        """ Initialize/Reinitialize the network """
        super()._initialize(**kwargs)
        # parameters that used in all layers
        # base parameters
        self.dims = kwargs.pop("dims", [1, 100, 1])
        self.biases = kwargs.pop("biases", None)
        self.weights = kwargs.pop("weights", 'uniform')

        # network dynamics parameters
        self.sparsity_masks = kwargs.pop("sparsity_masks", None)
        self.ei_masks = kwargs.pop("ei_masks", None)
        self.plasticity_masks = kwargs.pop("plasticity_masks", None)

        # temp storage
        self.preact_noise = kwargs.pop("preact_noise", 0)
        self.postact_noise = kwargs.pop("postact_noise", 0)

        # suppress warnings
        self.suppress_warnings = kwargs.pop("suppress_warnings", False)

        # check if all parameters meet the requirements
        if not self.suppress_warnings:
            self._handle_warnings(kwargs)
        self._check_parameters()
        rc_struct, out_struct = self._build_structures(kwargs)

        # layers
        self.recurrent_layer = RecurrentLayer(layer_struct=rc_struct)
        self.readout_layer = LinearLayer(layer_struct=out_struct)

    @property
    def layers(self):
        layer_list = [
            self.recurrent_layer.input_layer, 
            self.recurrent_layer.hidden_layer, 
            self.readout_layer
        ]
        return layer_list

    def _handle_warnings(self, kwargs):
        """ Handle deprecated parameters """
        slevel = 5 # output the warning at the level where kwargs is passed
        # check if there is any deprecated parameter
        if 'input_dim' in kwargs or 'output_dim' in kwargs or 'hidden_size' in kwargs:
            if 'dims' not in kwargs:
                self.dims = [kwargs.pop("input_dim", 1), kwargs.pop("hidden_size", 100), kwargs.pop("output_dim", 1)]
            warnings.warn("input_dim, output_dim, hidden_size are deprecated. Use dims instead.", UserWarning, stacklevel=slevel)
        if 'ei_balance' in kwargs:
            warnings.warn("ei_balance is deprecated. No ei_balance specification is needed.", UserWarning, stacklevel=slevel)
        if 'allow_negative' in kwargs:
            warnings.warn("allow_negative is deprecated. No allow_negative specification is needed.", UserWarning, stacklevel=slevel)
        if 'use_dale' in kwargs:
            warnings.warn("use_dale is deprecated. Use ei_masks instead. No Dale's law is applied.", UserWarning, stacklevel=slevel)
        if 'new_synapses' in kwargs:
            warnings.warn("new_synapses is deprecated. Use sparsity_masks instead. No synapse constraint is applied.", UserWarning, stacklevel=slevel)
        if 'learnable' in kwargs:
            warnings.warn("learnable is deprecated. Use `plasticity_masks` instead.\n"
                "   If you wish to define the learning behavior of the weights, generate a matrix "
                "with the desired learning rate and pass it to the model via `plasticity_masks`.", UserWarning, stacklevel=slevel)
        if 'positivity_constraints' in kwargs:
            warnings.warn("positivity_constraints is deprecated. Use `ei_masks` instead.\n"
                "   If you wish to constraint the positivity of the weights, generate a matrix with the "
                "desired positivity/netagivity and pass it to the model via `ei_masks`.", UserWarning, stacklevel=slevel)
        if 'sparsity_constraints' in kwargs:
            warnings.warn("sparsity_constraints is deprecated. Use sparsity_masks instead.\n"
                "   If you wish to constraint the sparsity of the weights, mask the element that you wish "
                "to constrain to zero and pass the mask to the model via `sparsity_masks`.", UserWarning, stacklevel=slevel)
        if 'scaling' in kwargs:
            warnings.warn("scaling is deprecated. Use `weights` instead.\n"
                "   If you wish to scale the weights, generate the weights with the desired "
                "scaling and pass them to the model via `weights`.", UserWarning, stacklevel=slevel)
        if 'self_connections' in kwargs:
            warnings.warn("self_connections is deprecated. Use sparsity_masks instead.\n"
                "   If you wish to constraint the sparsity of the weights, mask the element that you wish, "
                "i.e. the diagonal elements, to constrain to zero and pass the mask to the model via `sparsity_masks`.", UserWarning, stacklevel=slevel)
        if 'layer_distributions' in kwargs:
            warnings.warn("layer_distributions is deprecated. Use `weights` instead.\n"
                "   The parameter `weights` inherits the functionality of `layer_distributions`."
                "Simply pass a list of distributions to `weights` and the model will generate the weights accordingly.", UserWarning, stacklevel=slevel)
        if 'layer_biases' in kwargs:
            warnings.warn("layer_biases is deprecated. Use `biases` instead.\n"
                "   The parameter `biases` inherits the functionality of `layer_biases`."
                "Simply pass a list of biases to `biases` and the model will generate the biases accordingly.", UserWarning, stacklevel=slevel)
        if 'layer_masks' in kwargs:
            warnings.warn("layer_masks is deprecated. Use `sparsity_masks`, `ei_masks`, and `plasticity_masks` instead.\n"
                "   The parameter `sparsity_masks`, `ei_masks`, and `plasticity_masks` inherits the functionality of `layer_masks`."
                "Simply pass a list of masks to `sparsity_masks`, `ei_masks`, and `plasticity_masks` and the model will generate the masks accordingly.", UserWarning, stacklevel=slevel)
        
        for key in kwargs:
            if not key in param_list and not key in dep_param_list:
                print("unrecognized parameter: {}".format(key))

    def _check_masks(self, param, param_type, dims):
        """ General function to check different parameter types. """
        target_dim = [(dims[0], dims[1]), (dims[1], dims[1]), (dims[1], dims[2])]
        target_dim_biases = [dims[1], dims[1], dims[2]]
 
        # Handle None cases
        if param is None:
            if param_type in ["ei_masks", "sparsity_masks", "plasticity_masks", "biases"]:
                param = [None] * 3
            else: raise ValueError(f"{param_type} cannot be None when param_type is {param_type}") # weights
        elif param is not None and type(param) != list and param_type in ["weights"]:
            param = [param] * 3

        if type(param) != list:
            if param_type in ["ei_masks", "sparsity_masks", "plasticity_masks"]:
                raise ValueError(f"{param_type} must be a list of 3 values")
            else:
                param = [param] * 3
        if len(param) != 3:
            raise ValueError(f"{param_type} is/can not be broadcasted to a list of length 3")

        # param_type are all legal because it is passed by non-user code
        if param_type == "plasticity_masks": param = self._regularize_plas_masks(param, target_dim)
        else:
            # if its not plasticity_masks, then it must be a list of 3 values
            for i in range(3):
                if param[i] is not None:
                    if param_type in ["ei_masks", "sparsity_masks"]:
                        param[i] = self._check_array(param[i], param_type, target_dim[i], i)
                        if param_type == "ei_masks":
                            param[i] = torch.where(param[i] > 0, torch.tensor(1), torch.tensor(-1))
                        elif param_type == "sparsity_masks":
                            param[i] = torch.where(param[i] == 0, torch.tensor(0), torch.tensor(1))
                    elif param_type in ["weights", "biases"]:
                        self._check_distribution_or_array(param[i], param_type, target_dim_biases[i] if param_type == "biases" else target_dim[i], i)
        return param

    def _regularize_plas_masks(self, masks, target_dim):
        if any(mask is not None for mask in masks):
            min_plas, max_plas = [], []
            for mask in masks:
                if mask is not None:
                    min_plas.append(mask.min())
                    max_plas.append(mask.max())
            min_plas, max_plas = min(min_plas), max(max_plas)
            if min_plas != max_plas:
                params = []
                for i in range(3):
                    if masks[i] is None: params.append(torch.ones(target_dim[i]))
                    else:
                        _temp_mask = (masks[i] - min_plas) / (max_plas - min_plas)
                        params.append(self._check_array(_temp_mask, "plasticity_masks", target_dim[i], i))
                # check the total number of unique plasticity values
                plasticity_scales = torch.unique(torch.cat([param.flatten() for param in params]))
                if len(plasticity_scales) > 5:
                    raise ValueError("The number of unique plasticity values cannot be larger than 5")
                return params
        return [torch.ones(target_dim[i]) for i in range(3)]

    def _check_array(self, param, param_type, dim, index):
        if type(param) != torch.Tensor:
            if type(param) == np.ndarray:
                param = torch.from_numpy(param)
            else: raise ValueError(f"{param_type}[{index}] must be a numpy array or torch tensor")
        if param.shape != dim:
            raise ValueError(f"{param_type}[{index}] must be a numpy array of shape {dim}")
        return param

    def _check_distribution_or_array(self, param, param_type, dim, index):
        if type(param) == str:
            if param not in ['uniform', 'normal', 'zero']:
                raise ValueError(f"{param_type}[{index}] must be a string of 'uniform' or 'normal'")
        elif type(param) == torch.Tensor:
            # its already being converted to torch.Tensor, so no need to check np.ndarray case
            if param.shape != dim:
                raise ValueError(f"{param_type}[{index}] must be a numpy array of shape {dim}")
        else:
            raise ValueError(f"{param_type}[{index}] must be a string of 'uniform' or 'normal' \
                or a numpy array/torch tensor with shape {dim}")

    def _check_parameters(self):
        """ Check parameters """
        # check dims
        assert type(self.dims) == list, "dims must be a list"
        assert len(self.dims) == 3, "dims must be a list of length 3"
        for i in self.dims:
            assert type(i) == int, f"dims must be a list of integers, {i} is not an integer"
            assert i > 0, "dims must be a list of positive integers"

        # check ei_masks
        self.ei_masks = self._check_masks(self.ei_masks, "ei_masks", self.dims)
        # check weights
        self.weights = self._check_masks(self.weights, "weights", self.dims)
        # check biases
        self.biases = self._check_masks(self.biases, "biases", self.dims)
        # check sparsity_masks
        self.sparsity_masks = self._check_masks(self.sparsity_masks, "sparsity_masks", self.dims)
        # check plasticity_masks
        self.plasticity_masks = self._check_masks(self.plasticity_masks, "plasticity_masks", self.dims)

    def _build_structures(self, kwargs):
        """ Build structures """
        # build structures
        rc_struct = {
            "init_state": kwargs.pop("init_state", 'zero'),
            "activation": kwargs.pop("activation", "relu"),
            "dt": kwargs.pop("dt", 10),
            "tau": kwargs.pop("tau", 100),
            "preact_noise": self.preact_noise,
            "postact_noise": self.postact_noise,
            "in_struct": {
                "input_dim": self.dims[0],
                "output_dim": self.dims[1],
                "weights": self.weights[0],
                "biases": self.biases[0],
                "sparsity_mask": self.sparsity_masks[0],
                "ei_mask": self.ei_masks[0],
                "plasticity_mask": self.plasticity_masks[0],
            },
            "hid_struct": {
                "input_dim": self.dims[1],
                "output_dim": self.dims[1],
                "weights": self.weights[1],
                "biases": self.biases[1],
                "sparsity_mask": self.sparsity_masks[1],
                "ei_mask": self.ei_masks[1],
                "plasticity_mask": self.plasticity_masks[1],
            }
        }
        out_struct = {
            "input_dim": self.dims[1],
            "output_dim": self.dims[2],
            "weights": self.weights[2],
            "biases": self.biases[2],
            "sparsity_mask": self.sparsity_masks[2],
            "ei_mask": self.ei_masks[2],
            "plasticity_mask": self.plasticity_masks[2],
        }
        return rc_struct, out_struct
    # ======================================================================================

    # FORWARD
    # ======================================================================================    
    def to(self, device):
        """ Move the network to device """
        super().to(device)
        self.recurrent_layer.to(device)
        self.readout_layer.to(device)

    def forward(self, x):
        """ 
        Forwardly update network

        Inputs:
            - x: input, shape: (n_timesteps, batch_size, input_dim)
        """
        # skip constraints if the model is not in training mode
        if self.training:
            self._enforce_constraints()
        hidden_states = self.recurrent_layer(x)
        output = self.readout_layer(hidden_states.float())
        return output, [hidden_states]

    def train(self):
        """
        Set pre-activation and post-activation noise to the specified value
        and resume enforcing constraints
        """
        self.recurrent_layer.preact_noise = self.preact_noise
        self.recurrent_layer.postact_noise = self.postact_noise
        self.training = True

    def eval(self):
        """ 
        Set pre-activation and post-activation noise to zero 
        and pause enforcing constraints
        """
        self.recurrent_layer.preact_noise = 0
        self.recurrent_layer.postact_noise = 0
        self.training = False

    def apply_plasticity(self):
        """
        Apply plasticity to the weight gradient such that the weights representing the synapses
        will update at different rates according to the plasticity mask.
        """
        # no need to consider the case where plasticity_mask is None as 
        # it will be automatically converted to a tensor of ones in parameter initialization
        self.recurrent_layer.apply_plasticity()
        self.readout_layer.apply_plasticity()

    def _enforce_constraints(self):
        self.recurrent_layer.enforce_constraints()
        self.readout_layer.enforce_constraints()

    # HELPER FUNCTIONS
    # ======================================================================================
    def print_layers(self):
        """ Print the specs of each layer """
        self.recurrent_layer.print_layers()
        self.readout_layer.print_layers()

    def plot_layers(self):
        """ Plot the weights matrix and distribution of each layer """
        self.recurrent_layer.plot_layers()
        self.readout_layer.plot_layers()
    # ======================================================================================



import numpy as np
from .multi_area import MultiArea

class MultiAreaEI(MultiArea):
    """ Multi-area network with ei constraints mask """
    def __init__(self, **kwargs):
        """
        NOTE: On top of the keywards required for MultiArea:
        @kwarg exc_pct: percentage of excitatory neurons, default: 0.8
        @kwarg inter_area_connections: list of boolean, whether to have inter-area connections
            [exc_exc, exc_inh, inh_exc, inh_inh]
            default: [True, True, True, True]
        @kwarg inh_readout: whether to readout inhibitory neurons, default: True
        """
        super().__init__(**kwargs)
        # initialize parameters
        self.exc_pct = kwargs.get("exc_pct", 0.8)
        self.inter_area_connections = kwargs.get("inter_area_connections", [True, True, True, True])
        self.inh_readout = kwargs.get("inh_readout", True)
        # check parameters and generate mask
        self._check_parameters()
        self._generate_mask()

    def _check_parameters(self):
        super()._check_parameters()
        # check exc_pct
        assert 0 <= self.exc_pct <= 1, "exc_pct must be between 0 and 1"
        # check if inter_area_connections is list of 4 boolean
        assert isinstance(self.inter_area_connections, list) and len(self.inter_area_connections) == 4, \
            "inter_area_connections must be list of 4 boolean"
        for i in range(4):
            assert isinstance(self.inter_area_connections[i], bool), "inter_area_connections must be list of 4 boolean"

    def _generate_mask(self):
        """
        Generate the mask for the multi-area network
        """
        super()._generate_mask()
        self._generate_ei_assigment()
        self._masks_to_ei()

    def _generate_ei_assigment(self):
        """
        Generate the assignment of excitatory and inhibitory neurons
        """
        self.excitatory_neurons = np.zeros(self.n_areas, dtype=np.ndarray)
        self.inhibitory_neurons = np.zeros(self.n_areas, dtype=np.ndarray)

        for i in range(self.n_areas):
            area_i_size = len(self.node_assigment[i])
            n_exc = int(area_i_size * self.exc_pct)
            self.excitatory_neurons[i] = self.node_assigment[i][:n_exc]
            self.inhibitory_neurons[i] = self.node_assigment[i][n_exc:]

    def _masks_to_ei(self):
        """
        Convert the masks to ei masks
        """
        for i in range(self.n_areas):
            self.hidden_mask[:, self.inhibitory_neurons[i]] *= -1
            self.readout_mask[:, self.inhibitory_neurons[i]] *= -1

        # remove exc_exc connections between areas
        if not self.inter_area_connections[0]:
            for i in range(self.n_areas):
                for j in range(i+1, self.n_areas):
                    self.hidden_mask[np.ix_(self.excitatory_neurons[i], self.excitatory_neurons[j])] = 0
                    self.hidden_mask[np.ix_(self.excitatory_neurons[j], self.excitatory_neurons[i])] = 0

        # remove exc_inh connections between areas
        if not self.inter_area_connections[1]:
            for i in range(self.n_areas):
                for j in range(self.n_areas):
                    if i != j:
                        self.hidden_mask[np.ix_(self.inhibitory_neurons[i], self.excitatory_neurons[j])] = 0

        # remove inh_exc connections between areas
        if not self.inter_area_connections[2]:
            for i in range(self.n_areas):
                for j in range(self.n_areas):
                    if i != j:
                        self.hidden_mask[np.ix_(self.excitatory_neurons[i], self.inhibitory_neurons[j])] = 0

        # remove inh_inh connections between areas
        if not self.inter_area_connections[3]:
            for i in range(self.n_areas):
                for j in range(i+1, self.n_areas):
                    self.hidden_mask[np.ix_(self.inhibitory_neurons[i], self.inhibitory_neurons[j])] = 0
                    self.hidden_mask[np.ix_(self.inhibitory_neurons[j], self.inhibitory_neurons[i])] = 0

        # delete connections from inhibitory neurons to readout layers
        if not self.inh_readout:
            for i in range(self.n_areas):
                self.readout_mask[:, self.inhibitory_neurons[i]] = 0



import torch
import torch.nn as nn


class MLPLoss(nn.Module):
    def __init__(self, model, **kwargs):
        super().__init__()
        self.model = model
        self._init_losses(**kwargs)

    def _init_losses(self, **kwargs):
        """
        Initialize the loss functions
        """
        self.lambda_mse = kwargs.get("lambda_mse", 1)

        # the number of loss functions
        lambda_list, loss_list = [], []

        # init lambdas
        # TODO: weight constraints
        lambda_list.append(kwargs.get("lambda_fr", 0))
        lambda_list.append(kwargs.get("lambda_fr_sd", 0))
        # init loss functions
        loss_list.append(self._loss_fr)
        loss_list.append(self._loss_fr_sd)

        # save to self
        self.lambda_list = torch.tensor(lambda_list)
        self.loss_list = loss_list

    def _loss_fr(self, states, **kwargs):
        """ Compute the loss for firing rate """
        # return torch.sqrt(torch.square(states)).mean()
        loss = []
        for s in states:
            l = torch.pow(torch.mean(s, dim=(0, 1)), 2).mean()
            loss.append(l)
        return torch.stack(loss).mean()

    def _loss_fr_sd(self, states, **kwargs):
        """ Compute the loss for firing rate for each neuron in terms of SD """
        # return torch.sqrt(torch.square(states)).mean(dim=(0)).std()
        return torch.pow(torch.mean(states, dim=(0, 1)), 2).std()

    def forward(self, pred, label, **kwargs):
        """
        Compute the loss
        @param pred: size=(-1, batch_size, 2), predicted labels
        @param label: size=(-1, batch_size, 2), labels
        @param dur: duration of the trial
        """
        loss = [self.lambda_mse * torch.square(pred-label).mean()]
        for i in range(len(self.loss_list)):
            if self.lambda_list[i] == 0:
                continue
            else:
                loss.append(self.lambda_list[i]*self.loss_list[i](**kwargs))
        loss = torch.stack(loss)
        return loss.sum(), loss

    def to(self, device):
        """ Move to device """
        super().to(device)
        self.lambda_list = self.lambda_list.to(device)


from .multi_area import MultiArea
from .multi_area_ei import MultiAreaEI
from .random_input import RandomInput

if __name__ == "__main__":
    print(MultiAreaEI)
    print(MultiArea)
    print(RandomInput)



import torch
import torch.nn as nn

from nn4n.model import CTRNN

class RNNLoss(nn.Module):
    """
    Loss function for RNN

    Inputs:
        - model: the model, must be an nn4n.model.CTRNN

    Keyword Arguments:
        - lambda_mse: coefficient for the MSE loss, default: 1
        - lambda_in: coefficient for the input layer loss, default: 0
        - lambda_hid: coefficient for the hidden layer loss, default: 0
        - lambda_out: coefficient for the readout layer loss, default: 0
        - lambda_fr: coefficient for the overall firing rate loss, default: 0
        - lambda_fr_sd: coefficient for the standard deviation of firing rate 
            loss (to evenly distribute firing rate across neurons), default: 0
        - lambda_fr_cv: coefficient for the coefficient of variation of firing
            rate loss (to evenly distribute firing rate across neurons), default: 0
    
    Note that these firing rate does not automatically normalized to a similar magnitude
    """
    def __init__(self, model, **kwargs):
        super().__init__()
        self.model = model
        if type(self.model) != CTRNN:
            raise TypeError("model must be CTRNN")
        self._init_losses(**kwargs)

    def _init_losses(self, **kwargs):
        """
        Initialize the loss functions
        """
        self.lambda_mse = kwargs.get("lambda_mse", 1)

        # the number of loss functions
        lambda_list, loss_list = [], []

        # init lambdas
        lambda_list.append(kwargs.get("lambda_in", 0))
        lambda_list.append(kwargs.get("lambda_hid", 0))
        lambda_list.append(kwargs.get("lambda_out", 0))
        lambda_list.append(kwargs.get("lambda_fr", 0))
        lambda_list.append(kwargs.get("lambda_fr_sd", 0))
        lambda_list.append(kwargs.get("lambda_fr_cv", 0))
        # init loss functions
        loss_list.append(self._loss_in)
        loss_list.append(self._loss_hid)
        loss_list.append(self._loss_out)
        loss_list.append(self._loss_fr)
        loss_list.append(self._loss_fr_sd)
        loss_list.append(self._loss_fr_cv)

        # save to self
        self.lambda_list = lambda_list
        self.loss_list = loss_list

        # init constants
        n_in = self.model.recurrent_layer.input_layer.weight.shape[1]
        n_size = self.model.recurrent_layer.hidden_layer.weight.shape[0]
        n_out = self.model.readout_layer.weight.shape[0]
        self.n_in_dividend = n_in*n_size
        self.n_hid_dividend = n_size*n_size
        self.n_out_dividend = n_out*n_size

    def _loss_in(self, **kwargs):
        """ Compute the loss for InputLayer """
        return torch.norm(self.model.recurrent_layer.input_layer.weight, p='fro')**2/self.n_in_dividend

    def _loss_hid(self, **kwargs):
        """ Compute the loss for RecurrentLayer """
        return torch.norm(self.model.recurrent_layer.hidden_layer.weight, p='fro')**2/self.n_hid_dividend

    def _loss_out(self, **kwargs):
        """ Compute the loss for ReadoutLayer """
        return torch.norm(self.model.readout_layer.weight, p='fro')**2/self.n_out_dividend

    def _loss_fr(self, states, **kwargs):
        """ Compute the loss for firing rate """
        return torch.pow(torch.mean(states, dim=(0, 1)), 2).mean()

    def _loss_fr_sd(self, states, **kwargs):
        """ Compute the loss for firing rate for each neuron in terms of SD """
        return torch.pow(torch.mean(states, dim=(0, 1)), 2).std()

    def _loss_fr_cv(self, states, **kwargs):
        """ Compute the loss for firing rate for each neuron in terms of coefficient of variation """
        avg_fr = torch.sqrt(torch.square(states)).mean(dim=0)
        return avg_fr.std()/avg_fr.mean()

    def forward(self, pred, label, **kwargs):
        """
        Compute the loss

        Inputs:
            - pred: size=(-1, batch_size, 2), predicted labels
            - label: size=(-1, batch_size, 2), labels
        
        where -1 is the sequence length
        """
        loss = [self.lambda_mse * torch.square(pred-label).mean()]
        for i in range(len(self.loss_list)):
            if self.lambda_list[i] == 0:
                continue
            else:
                loss.append(self.lambda_list[i]*self.loss_list[i](**kwargs))
        loss = torch.stack(loss)
        return loss.sum(), loss



# created By: Zhaoze Wang
from setuptools import setup, find_packages
from pathlib import Path
this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

setup(
    name='nn4n',
    version='1.1.0',
    description='Neural Networks for Neuroscience Research',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Zhaoze Wang',
    license='MIT',

    packages=find_packages(),
    install_requires=[
        'numpy',
        'matplotlib',
        'torch',
        'IPython',
        'scipy',
    ],
    python_requires='>=3.10',
)



from .linear_layer import LinearLayer
from .hidden_layer import HiddenLayer
from .recurrent_layer import RecurrentLayer

if __name__ == '__main__':
    print(LinearLayer)
    print(HiddenLayer)
    print(RecurrentLayer)